@@ -36,7 +36,7 @@ from tqdm import tqdm
 from pandas import DataFrame
 from .helper_functions import _get_sofar_api, get_spotter_ids
 import numpy
-import os
+import roguewave.spotterapi.spotter_cache as spotter_cache
 
 # 2) Constants & Private Variables
 # =============================================================================
@@ -53,9 +53,6 @@ MAXIMUM_NUMBER_OF_WORKERS = 40
 # Number of retry attemps if a download fails.
 NUMBER_OF_RETRIES = 2
 
-# Spotter Cache
-CACHE_NAME = '~/roguewave/spotter_cache'
-
 
 # 3) Classes
 # =============================================================================
@@ -82,7 +79,7 @@ def get_spectrum(
         end_date: Union[datetime, int, float, str] = None,
         session: SofarApi = None,
         parallel_download=True,
-        cache = False
+        cache: bool = True
 ) -> Dict[str, List[WaveSpectrum1D]]:
     """
     Gets the requested frequency wave data for the spotter(s) in the given
@@ -94,14 +91,22 @@ def get_spectrum(
     :param start_date: ISO 8601 formatted date string, epoch or datetime.
                        If not included defaults to beginning of spotters
                        history
+
     :param end_date:   ISO 8601 formatted date string, epoch or datetime.
                        If not included defaults to end of spotter history
+
     :param session:    Active SofarApi session. If none is provided one will be
                        creatated automatically. This requires that an API key
                        is set in the environment.
+
     :param parallel_download: Use multiple requests to the Api to speed up
                         retrieving data. Only useful for large requests.
 
+    :param cache: Cache requests. If True, returned data will be stored in
+                        a file Cache on disk, and repeated calls with the
+                        same arguments will use locally cached data. The cache
+                        is a FileCache with a maximum of 2GB by default.
+
     :return: Data as a dictornary with spotter_id's as keys, and for each
     corresponding value a List that for each returned timestamp contains a
     WaveSpectrum1D object.
@@ -117,6 +122,7 @@ def get_spectrum(
         include_surface_temp_data=False,
         session=session,
         parallel_download=parallel_download,
+        cache=cache
     )
     out = {}
     for key in data:
@@ -133,7 +139,9 @@ def get_bulk_wave_data(
         end_date: Union[datetime, int, float, str] = None,
         session: SofarApi = None,
         parallel_download: bool = True,
-        bulk_data_as_dataframe: bool = True
+        bulk_data_as_dataframe: bool = True,
+        cache: bool = True,
+        convert_to_sofar_model_names = False
 ) -> Union[Dict[str, List[WaveBulkData]], Dict[str, DataFrame]]:
     """
     Gets the requested bulk wave data for the spotter(s) in the given interval
@@ -144,17 +152,26 @@ def get_bulk_wave_data(
     :param start_date: ISO 8601 formatted date string, epoch or datetime.
                        If not included defaults to beginning of spotters
                        history
+
     :param end_date:   ISO 8601 formatted date string, epoch or datetime.
                        If not included defaults to end of spotter history
+
     :param session:    Active SofarApi session. If none is provided one will be
                        creatated automatically. This requires that an API key
                        is set in the environment.
+
     :param parallel_download: Use multiple requests to the Api to speed up
                         retrieving data. Only useful for large requests.
+
     :param bulk_data_as_dataframe: return bulk data as a dataframe per Spotter
                        instead of a list of BulkWaveVariable objects.
                        Defaults to yes- only set to false for dev purposes.
 
+    :param cache: Cache requests. If True, returned data will be stored in
+                        a file Cache on disk, and repeated calls with the
+                        same arguments will use locally cached data. The cache
+                        is a FileCache with a maximum of 2GB by default.
+
     :return: Data as a dictornary with spotter_id's as keys, and for each
     corresponding value a dataframe containing the output.
     """
@@ -168,7 +185,9 @@ def get_bulk_wave_data(
         include_surface_temp_data=False,
         session=session,
         parallel_download=parallel_download,
-        bulk_data_as_dataframe=bulk_data_as_dataframe
+        bulk_data_as_dataframe=bulk_data_as_dataframe,
+        cache=cache,
+        convert_to_sofar_model_names=convert_to_sofar_model_names
     )
     out = {}
     for key in data:
@@ -191,7 +210,8 @@ def get_data(
         session: SofarApi = None,
         parallel_download=True,
         bulk_data_as_dataframe=True,
-        cache=False
+        cache=True,
+        convert_to_sofar_model_names=False
 ) -> Dict[str, Dict[str, Union[list[WaveSpectrum1D],
                                list[WaveBulkData], DataFrame]]]:
     """
@@ -200,25 +220,42 @@ def get_data(
     :param spotter_ids: Can be either 1) a List of spotter_ids or 2) a single
     Spotter_id.
 
-    :param include_frequency_data:
-    :param include_waves:
-    :param include_wind:
-    :param include_surface_temp_data:
-    :param include_barometer_data:
     :param start_date: ISO 8601 formatted date string, epoch or datetime.
                        If not included defaults to beginning of spotters
                        history
+
     :param end_date:   ISO 8601 formatted date string, epoch or datetime.
                        If not included defaults to end of spotter history
+
+    :param include_frequency_data: set to True to return spectral data
+        (if available)
+
+    :param include_waves: set to True to return wave data (if available)
+
+    :param include_wind: set to True to return wind data (if available)
+
+    :param include_surface_temp_data: set to True to return SST data
+        (if available)
+
+    :param include_barometer_data: set to True to return barometer data
+        (if available)
+
     :param session:    Active SofarApi session. If none is provided one will be
                        created automatically. This requires that an API key is
                        set in the environment.
+
     :param parallel_download: Use multiple requests to the Api to speed up
                        retrieving data. Only useful for large requests.
+
     :param bulk_data_as_dataframe: return bulk data as a dataframe per Spotter
                        instead of a list of BulkWaveVariable objects. Defaults
                        to yes- only set to false for dev purposes.
 
+    :param cache: Cache requests. If True, returned data will be stored in
+                        a file Cache on disk, and repeated calls with the
+                        same arguments will use locally cached data. The cache
+                        is a FileCache with a maximum of 2GB by default.
+
     :return: Data as a dictornary with spotter_id's as keys, and for each
     corresponding value a dataframe containing the output.
     """
@@ -238,21 +275,37 @@ def get_data(
     if not isinstance(spotter_ids, list):
         spotter_ids = [spotter_ids]
 
-    if session is None:
-        session = _get_sofar_api()
+    print(f"Get spotter data: retrieving data for {len(spotter_ids)} Spotters")
+    if cache and end_date < datetime.now(tz=timezone.utc):
+        # Use the cache _only_ for requests that concern the past. In real-time
+        # things may change.
 
-    def worker(_spotter_id):
-        return _download_data(_spotter_id, session, variables_to_include,
-                              start_date, end_date, bulk_data_as_dataframe)
-
-    if parallel_download:
-        with ThreadPool(processes=MAXIMUM_NUMBER_OF_WORKERS) as pool:
-            out = list(
-                tqdm(pool.imap(worker, spotter_ids),
-                     total=len(spotter_ids)))
+        out = spotter_cache.get_data(
+            spotter_ids,
+            session,
+            _download_data,
+            variables_to_include=variables_to_include,
+            start_date=start_date,
+            end_date=end_date,
+            bulk_data_as_dataframe=bulk_data_as_dataframe,
+            convert_to_sofar_model_names=convert_to_sofar_model_names)
     else:
-        out = list(
-            tqdm(map(worker, spotter_ids), total=len(spotter_ids)))
+        if session is None:
+            session = _get_sofar_api()
+
+        def worker(_spotter_id):
+            return _download_data(_spotter_id, session, variables_to_include,
+                                  start_date, end_date, bulk_data_as_dataframe,
+                                  convert_to_sofar_model_names)
+
+        if parallel_download:
+            with ThreadPool(processes=MAXIMUM_NUMBER_OF_WORKERS) as pool:
+                out = list(
+                    tqdm(pool.imap(worker, spotter_ids),
+                         total=len(spotter_ids)))
+        else:
+            out = list(
+                tqdm(map(worker, spotter_ids), total=len(spotter_ids)))
 
     data = {}
     for spotter_id, spotter_data in zip(spotter_ids, out):
@@ -273,6 +326,7 @@ def search_circle(
         center_lat_lon: Tuple,
         radius: float,
         session: SofarApi = None,
+        cache=False
 ):
     """
     Search for all Spotters that have data available within the give spatio-
@@ -291,7 +345,21 @@ def search_circle(
 
     geometry = {'type': 'circle', 'points': center_lat_lon, 'radius': radius}
 
-    return _search(start_date, end_date, geometry, session)
+    if session is None:
+        session = _get_sofar_api()
+
+    print("Get Spotter data: retrieving all data from spatio-temporal region")
+    if cache:
+        return spotter_cache.get_data_search(
+            handler=_search,
+            session=session,
+            start_date=start_date,
+            end_date=end_date,
+            geometry=geometry
+        )
+    else:
+        #
+        return _search(start_date, end_date, geometry, session)
 
 
 # -----------------------------------------------------------------------------
@@ -302,6 +370,7 @@ def search_rectangle(
         end_date: Union[datetime, str],
         bounding_box,
         session: SofarApi = None,
+        cache=False
 ):
     """
     Search for all Spotters that have data available within the give spatio-
@@ -323,7 +392,21 @@ def search_rectangle(
                 'points': bounding_box,
                 'radius': None}
 
-    return _search(start_date, end_date, geometry, session)
+    if session is None:
+        session = _get_sofar_api()
+
+    print("Get Spotter data: retrieving all data from spatio-temporal region")
+    if cache:
+        return spotter_cache.get_data_search(
+            _search,
+            session,
+            start_date=start_date,
+            end_date=end_date,
+            geometry=geometry
+        )
+    else:
+        #
+        return _search(start_date, end_date, geometry, session)
 
 
 # -----------------------------------------------------------------------------
@@ -338,7 +421,8 @@ def _download_data(
         start_date: Union[datetime, str, int, float] = None,
         end_date: Union[datetime, str, int, float] = None,
         bulk_data_as_dataframe: bool = True,
-        limit: int = None) -> \
+        limit: int = None,
+        convert_to_sofar_model_names=False) -> \
         Dict[str, Union[List[WaveSpectrum1D], List[WaveBulkData]]]:
     """
     Function that downloads data from the API for the requested Spotter
@@ -436,7 +520,7 @@ def _download_data(
                 if key == 'frequencyData':
                     # We cannot convert the list of wavespectra to a dataframe.
                     continue
-                data[key] = as_dataframe(data[key])
+                data[key] = as_dataframe(data[key],convert_to_sofar_model_names)
 
     return data
 
@@ -565,7 +649,7 @@ def _search(start_date: Union[datetime, str],
             session: SofarApi = None,
             variables_to_include: VariablesToInclude = None,
             page_size=500,
-            bulk_data_as_dataframe: bool = True
+            bulk_data_as_dataframe: bool = True,
             ):
     if session is None:
         session = _get_sofar_api()
@@ -668,7 +752,8 @@ def _none_filter(data: List[Union[WaveSpectrum1D, WaveBulkData]]):
     return list(
         filter(
             lambda x: (x['latitude'] is not None) and (
-                x['longitude'] is not None) and (x['timestamp'] is not None),
+                    x['longitude'] is not None) and (
+                                  x['timestamp'] is not None),
             data
         )
     )
@@ -703,33 +788,3 @@ def _get_class(key, data) -> Union[MetoceanData, WaveSpectrum1D]:
     else:
         raise Exception('Unknown variable')
 # -----------------------------------------------------------------------------
-
-
-def spotter_cache_uri( request_type,
-                       spotter_id,
-                       start_date: Union[datetime, int, float, str] = None,
-                       end_date: Union[datetime, int, float, str] = None,
-                       include_frequency_data=False,
-                       include_waves=True,
-                       include_wind=False,
-                       include_barometer_data=False,
-                       include_surface_temp_data=False,
-                       bulk_data_as_dataframe=True):
-
-    if end_date > datetime.now():
-        return None
-
-    uri = os.path.join( f"{os.getcwd()}",
-          f"{request_type}"
-          f"{spotter_id}"
-          f"{start_date.isoformat()}"
-          f"{end_date.isoformat()}"
-          f"{include_frequency_data}"
-          f"{include_waves}"
-          f"{include_wind}"
-          f"{include_barometer_data}"
-          f"{include_surface_temp_data}"
-          f"{bulk_data_as_dataframe}"
-          f".json"
-    )
-    return uri
\ No newline at end of file
